/*******************************************************************
HILBEX
version 0.1, 1 Oct 2016

This file is part of the software package Hilbex by Herman
Haverkort. Use and distribution of this software is subject to
restrictions. Consult readme.txt for details, or download it from
http://spacefillingcurves.net/.
*******************************************************************/

void Filter::parse(const string& specification, const Filter& commonFilter)
{
  prefix =     commonFilter.prefix;
  properties = commonFilter.properties;
  stringstream s(specification);
  char c; s >> std::ws; c = s.peek();
  while (!s.eof())
  {
    bool confirm = true;
    switch(c)
    {
      case 'C': case 'L': case 'N': case 'S': case 'X': case 'Y':
        s >> prefix;        
        break;        
      case '(': case '<':      
        traversal = LimitedTraversal(s);
        break;
      case '-': case '+':
        if (c == '-') confirm = false;
        s >> c; // skip the -/+ and proceed to the following word:
      default:
        string word;
        s >> word;
        if (confirm) properties.confirm(word); else properties.reject(word);
        break;
    }
    s >> std::ws; c = s.peek();    
  }

  // check syntax of prefix; if it's a short form, unpack it
  bool shortForm = (
    prefix.length() > 6 && prefix.length() <= 8 && string("cfgrstv").find(prefix[6]) == string::npos);
  if (prefix.length() > 16) 
  { 
    cerr << "Specified name or prefix \"" << prefix << "\" is too long." << endl;
    properties.confirm(Invalid);
  }
  else if (
    (prefix.length() > 13 && prefix[13] != '.') ||
    (prefix.length() > 10 && prefix[10] != '.') ||
    (prefix.length() >  7 && prefix[7]  != '.' && !shortForm) ||
    (prefix.length() >  4 && prefix[4]  != '.'))
  {
    cerr << "Specified name or prefix \"" << prefix << "\" is not in the right format." << endl;
    properties.confirm(Invalid);
  }
  if (shortForm) 
  {
    string longPrefix = prefix.substr(0, 6) + prefix[5] + '.' + prefix[6] + prefix[6] + ".II";
    if (prefix.length() == 8) { longPrefix += '.'; longPrefix += prefix[7]; longPrefix += prefix[7]; }
    prefix = longPrefix;
    properties.confirm(Symmetric);
  }
  
  // check if filter is consistent with default
  int commonPrefix = std::min(prefix.length(), commonFilter.prefix.length());
  if (prefix.substr(0, commonPrefix) != commonFilter.prefix.substr(0, commonPrefix))
  {
    cerr << "Specified name or prefix \"" << prefix << "\" conflicts with specified common name or prefix \"" << commonFilter.prefix << "\"." << endl;
    properties.confirm(Invalid);
  }
  if (prefix.length() < commonFilter.prefix.length()) prefix = commonFilter.prefix;
  commonPrefix = std::min(prefix.length(), traversal.name().length());
  if (prefix.substr(0, commonPrefix) != traversal.name().substr(0, commonPrefix)) 
    properties.confirm(Invalid);
  else if (prefix.length() < traversal.name().length()) 
    prefix = traversal.name();
  
  for (int i = 0; i < (int) Invalid; ++i)
  {
    if ((properties.is((Property) i) && commonFilter.properties.isNot((Property) i)) ||
        (properties.isNot((Property) i) && commonFilter.properties.is((Property) i)))
    {
      cerr << "Specified properties are inconsistent with specified common properties." << endl;
      properties.confirm(Invalid);
      break;
    }
    if (traversal.requires((Property) i))
    {
      if (properties.isNot((Property) i)) { properties.confirm(Invalid); break; }
      properties.confirm((Property) i);
    }
    else if (traversal.forbids((Property) i))
    {
      if (properties.is((Property) i)) { properties.confirm(Invalid); break; }
      properties.reject((Property) i);
    }
  }
  
  // gate configurations by gate type are mutually exclusive 
  int nrRequiredGateConfigurations = 0; int nrForbiddenGateConfigurations = 0;
  for (int i = (int) VertexGated; i <= (int) FacetGated; ++i) 
  {
    if (properties.is((Property) i)) ++nrRequiredGateConfigurations;
    if (properties.isNot((Property) i)) ++nrForbiddenGateConfigurations;
  }
  if (nrForbiddenGateConfigurations == 5) properties.confirm(Invalid);
  if (nrRequiredGateConfigurations > 1)
  {
    cerr << "Inconsistent gate configurations specified" << endl;
    properties.confirm(Invalid);
  }
  else if (nrRequiredGateConfigurations == 1)
    for (int i = (int) VertexGated; i <= (int) FacetGated; ++i) 
      if (!properties.is((Property) i)) properties.reject((Property) i);

  // gate configurations by relation of the gates are mutually exclusive 
  nrRequiredGateConfigurations = 0; nrForbiddenGateConfigurations = 0;
  for (int i = (int) EdgeCrossing; i <= (int) CubeCrossing; ++i) 
  {
    if (properties.is((Property) i)) ++nrRequiredGateConfigurations;
    if (properties.isNot((Property) i)) ++nrForbiddenGateConfigurations;
  }
  if (nrForbiddenGateConfigurations == 3) properties.confirm(Invalid);
  if (nrRequiredGateConfigurations > 1)
  {
    cerr << "Inconsistent gate configurations specified" << endl;
    properties.confirm(Invalid);
  }
  else if (nrRequiredGateConfigurations == 1)
    for (int i = (int) EdgeCrossing; i <= (int) CubeCrossing; ++i) 
      if (!properties.is((Property) i)) properties.reject((Property) i);
      
  // by definition, pattern-istropic curves cannot be coordinate-shifting and standing  
  // we catch this here to avoid unnecessarily generating all curves
  if (properties.is(PatternIsotropic))
  {
    if (properties.is(CoordinateShifting) || properties.is(Standing))
    {
      cerr << "Specified properties are inconsistent: coordinate-shifting or standing curves cannot be pattern-isotropic." << endl;
      properties.confirm(Invalid);
    }
  }
}

Filter::Filter() {};

Filter::Filter(const string& specification, const Filter& commonFilter)
{
  parse(specification, commonFilter);
}

Filter::Filter(istream& specification, const Filter& commonFilter)
{
  string line;
  getline(specification, line);
  parse(line, commonFilter);
}

bool Filter::consistent() const
{
  return !properties.is(Invalid);
}

bool Filter::requires(Property p) const
{
  return properties.is(p); 
}

bool Filter::forbids(Property p) const
{
  return properties.isNot(p);
}

bool Filter::matchTraversal(const RecursiveTraversal& r, Transformation& t) const
{
  return traversal.match(r, t);
}

string Filter::name() const
{
  return prefix;
}


