/*******************************************************************
HILBEX
version 0.1, 1 Oct 2016

This file is part of the software package Hilbex by Herman
Haverkort. Use and distribution of this software is subject to
restrictions. Consult readme.txt for details, or download it from
http://spacefillingcurves.net/.
*******************************************************************/

#ifndef HILBERT3DSUITE_GEOMETRY
#define HILBERT3DSUITE_GEOMETRY

#include <iostream>
using std::istream;

const int d = 3;

class FPt;
class NPt;
class OPt;
class GPt;

class FPt
// Points with floating-point coordinates, used for povray output
{
  private:
    double x[3];
  public:
    FPt(const double x0 = 0, const double x1 = 0, const double x2 = 0);
    double& operator[](const int i);
    double operator[](const int i) const;
};

ostream& operator<<(ostream& o, const FPt& p);

class NPt
// Points with non-negative integer coordinates, used for input and output of approximating curves
{
  private: 
    int x[3];
  public:
    NPt(const int x0 = 0, const int x1 = 0, const int x2 = 0);
    int& operator[](const int i);
    int operator[](const int i) const;
    FPt povRay(const double gap, const int order) const;
};

ostream& operator<<(ostream& o, const NPt& p);

class OPt
// Used for points with odd coordinates in units of 1/2 (between -2^k units and 2^k units for k-th order approximating curves),
// and for translation vectors for such points.
{
  private: 
    int x[3];
  public:
    OPt(const int x0 = 0, const int x1 = 0, const int x2 = 0);
    int& operator[](const int i);
    int operator[](const int i) const;
    operator GPt() const;
};

OPt convert(const NPt& p, const int order = 1);
NPt convert(const OPt& p, const int order = 1);

class GPt
// Points with coordinates between -6 and 6 units of 1/6th, for gate locations
{
  private: 
    int x[3];
  public:
    GPt(const int x0 = 0, const int x1 = 0, const int x2 = 0);
    int& operator[](const int i);
    int operator[](const int i) const;
};

ostream& operator<<(ostream& o, const GPt& p);

template <class P> P operator+(const P& p, const P& q);
template <class P> P operator-(const P& p, const P& q);
template <class P> P operator*(const P& p, const int s);
template <class P> P operator*(const P& p, const double s);
template <class P> P operator/(const P& p, const int s);
template <class P> bool operator==(const P& p, const int s);
template <class P> bool operator!=(const P& p, const int s);
template <class P> bool coplanar(const P& p, const P& q);
template <class P> bool collinear(const P& p, const P& q);
template <class P> int sqDistance(const P& p, const P& q);

class GeometricTransformation
{
  private:
    int invperm[3];
    int invsign[3];
  public:
    GeometricTransformation(const int p0, const int p1, const int p2);
    GeometricTransformation(const int id = 0);
    GeometricTransformation(const GeometricTransformation& t);
    GeometricTransformation(istream& s);
    operator int() const;
    bool isReflection() const;
    GeometricTransformation inverse() const;
    OPt operator()(const OPt& p) const;
    GPt operator()(const GPt& p) const;
    GeometricTransformation operator*(const GeometricTransformation& t) const;
    bool operator==(const GeometricTransformation& t) const;
    friend ostream& operator<<(ostream& o, const GeometricTransformation& t);
};

ostream& operator<<(ostream& o, const GeometricTransformation& t);

class Transformation
{
  private:
    void parse(istream& s);
    
  public: 
    GeometricTransformation geometry;
    int direction; // 0 is forward; 7 is reversed
    Transformation(const GeometricTransformation& theGeometry = GeometricTransformation(0), const int theDirection = 0);
    Transformation(const int id);
    Transformation(istream& s);
    Transformation(const string& s);
    Transformation inverse() const;
    Transformation operator*(const Transformation& t) const;
    friend ostream& operator<<(ostream& o, const Transformation& t);
};
  
ostream& operator<<(ostream& o, const Transformation& t);

class Symmetry: public Transformation
{
  public:
    char symbol;
    Symmetry(const char theSymbol, const int p0, const int p1, const int p2);
};

class Symmetries: public vector<Symmetry>
{
  public:
    Symmetries();
};

Symmetries transformations;


#endif

